@@ -40,7 +40,7 @@ int handleDNSCryptQuery(char* packet, uint16_t len, std::shared_ptr<DNSCryptQuer
   }
 
   if(*decryptedQueryLen < static_cast<uint16_t>(sizeof(struct dnsheader))) {
-    g_stats.nonCompliantQueries++;
+    ++g_stats.nonCompliantQueries;
     return false;
   }
 
