@@ -12,6 +12,8 @@ StatNode::Stat StatNode::print(unsigned int depth, Stat newstat, bool silent) co
   childstat.nxdomains += s.nxdomains;
   childstat.servfails += s.servfails;
   childstat.drops += s.drops;
+  childstat.bytes += s.bytes;
+
   if(children.size()>1024 && !silent) {
     cout<<string(depth, ' ')<<name<<": too many to print"<<endl;
   }
@@ -23,7 +25,8 @@ StatNode::Stat StatNode::print(unsigned int depth, Stat newstat, bool silent) co
       childstat.noerrors<<" noerrors, "<< 
       childstat.nxdomains<<" nxdomains, "<< 
       childstat.servfails<<" servfails, "<< 
-      childstat.drops<<" drops"<<endl;
+      childstat.drops<<" drops, "<<
+      childstat.bytes<<" bytes"<<endl;
 
   newstat+=childstat;
 
@@ -39,11 +42,11 @@ void  StatNode::visit(visitor_t visitor, Stat &newstat, unsigned int depth) cons
   childstat.nxdomains += s.nxdomains;
   childstat.servfails += s.servfails;
   childstat.drops += s.drops;
+  childstat.bytes += s.bytes;
   childstat.remotes = s.remotes;
   
   Stat selfstat(childstat);
 
-
   for(const children_t::value_type& child :  children) {
     child.second.visit(visitor, childstat, depth+8);
   }
@@ -54,18 +57,15 @@ void  StatNode::visit(visitor_t visitor, Stat &newstat, unsigned int depth) cons
 }
 
 
-void StatNode::submit(const DNSName& domain, int rcode, const ComboAddress& remote)
+void StatNode::submit(const DNSName& domain, int rcode, unsigned int bytes, boost::optional<const ComboAddress&> remote)
 {
   //  cerr<<"FIRST submit called on '"<<domain<<"'"<<endl;
-  vector<string> tmp = domain.getRawLabels();
+  std::vector<string> tmp = domain.getRawLabels();
   if(tmp.empty())
     return;
 
-  deque<string> parts;
-  for(auto const i : tmp) {
-    parts.push_back(i);
-  }
-  children[parts.back()].submit(parts, "", rcode, remote, 1);
+  auto last = tmp.end() - 1;
+  children[*last].submit(last, tmp.begin(), "", rcode, bytes, remote, 1);
 }
 
 /* www.powerdns.com. -> 
@@ -75,29 +75,29 @@ void StatNode::submit(const DNSName& domain, int rcode, const ComboAddress& remo
    www.powerdns.com. 
 */
 
-void StatNode::submit(deque<string>& labels, const std::string& domain, int rcode, const ComboAddress& remote, unsigned int count)
+void StatNode::submit(std::vector<string>::const_iterator end, std::vector<string>::const_iterator begin, const std::string& domain, int rcode, unsigned int bytes, boost::optional<const ComboAddress&> remote, unsigned int count)
 {
-  if(labels.empty())
-    return;
   //  cerr<<"Submit called for domain='"<<domain<<"': ";
   //  for(const std::string& n :  labels) 
   //    cerr<<n<<".";
   //  cerr<<endl;
   if(name.empty()) {
 
-    name=labels.back();
+    name=*end;
     //    cerr<<"Set short name to '"<<name<<"'"<<endl;
   }
-  else 
-    ; //    cerr<<"Short name was already set to '"<<name<<"'"<<endl;
+  else {
+    //    cerr<<"Short name was already set to '"<<name<<"'"<<endl;
+  }
 
-  if(labels.size()==1) {
+  if(end == begin) {
     if (fullname.empty()) {
       fullname=name+"."+domain;
       labelsCount = count;
     }
     //    cerr<<"Hit the end, set our fullname to '"<<fullname<<"'"<<endl<<endl;
     s.queries++;
+    s.bytes += bytes;
     if(rcode<0)
       s.drops++;
     else if(rcode==0)
@@ -106,7 +106,10 @@ void StatNode::submit(deque<string>& labels, const std::string& domain, int rcod
       s.servfails++;
     else if(rcode==3)
       s.nxdomains++;
-    s.remotes[remote]++;
+
+    if (remote) {
+      s.remotes[*remote]++;
+    }
   }
   else {
     if (fullname.empty()) {
@@ -114,8 +117,8 @@ void StatNode::submit(deque<string>& labels, const std::string& domain, int rcod
       labelsCount = count;
     }
     //    cerr<<"Not yet end, set our fullname to '"<<fullname<<"', recursing"<<endl;
-    labels.pop_back();
-    children[labels.back()].submit(labels, fullname, rcode, remote, count+1);
+    --end;
+    children[*end].submit(end, begin, fullname, rcode, bytes, remote, count+1);
   }
 }
 
